/********************************************************************
 * updated on 08-25-2023
 * Peng Zeng (Auburn University)
 * 
 * The codes were updated due to the following changes of R. 
 * 1. R changed BLAS.h (refer to Section 6.6.1 Fortran character strings)
 * 2. R removed DOUBLE_EPS in R_ext/Constants.h. Define it explicitly.
 ********************************************************************/

/********************************************************************
 * C functions for sim-lasso
 * Peng Zeng (Auburn University)
 * 11-27-2010
 *
 * Reference: 
 *   Peng Zeng, Tianhong He, and Yu Zhu (2010). A lasso-type approach 
 *       for estimation and variable selection in single-index model.
 *       Journal of Computational and Graphical Statistics, accepted.
 ********************************************************************/

#ifndef  USE_FC_LEN_T
# define USE_FC_LEN_T
#endif
#include <Rconfig.h>
#include <R_ext/BLAS.h>
#include <R_ext/Lapack.h>
#ifndef FCONE
# define FCONE
#endif

#include <R.h>
#include <Rmath.h>

#define DOUBLE_EPS (2.220446e-16); 

const int ZERO = 0;
const int ONE  = 1;

const double dZERO =  0.0;
const double dONE  =  1.0;
const double dnONE = -1.0;

/********************************************************************
 * sum(x)
 ********************************************************************/

double sum(const int *n, const double *x)
{
  double ans = 0.0;
  for(int i = 0; i < (*n); i++) ans += x[i];
  return(ans);
}


/********************************************************************
 * sum(x * y)
 ********************************************************************/

double sum2(const int *n, const double *x, const double *y)
{
  double ans = 0.0;
  for(int i = 0; i < (*n); i++) ans += (x[i] * y[i]);
  return(ans);
}


/********************************************************************
 * sum(x * y * z)
 ********************************************************************/

double sum3(const int *n, 
     const double *x, const double *y, const double *z)
{
  double ans = 0.0;
  for(int i = 0; i < (*n); i++) ans += (x[i] * y[i] * z[i]);
  return(ans);
}


/********************************************************************
 * sum( (x - x0) * (y - y0) * z )
 ********************************************************************/

double sum3Ctr(const int *n, 
     const double *x, const double *x0, 
     const double *y, const double *y0, 
     const double *z)
{
  double ans = 0.0;
  for(int i = 0; i < (*n); i++) 
    ans += ( (x[i] - (*x0)) * (y[i] - (*y0)) * z[i] );
  return(ans);
}


/********************************************************************
 * sum(abs(x))
 ********************************************************************/

double sumabs(const int *n, const double *x)
{
  double ans = 0.0;
  for(int i = 0; i < (*n); i++)  ans += fabs(x[i]);
  return(ans);
}


/********************************************************************
 * sum(x * x)
 ********************************************************************/

double sumsqx(const int *n, const double *x, const int *xinc)
{
  double ans = 0.0;
  for(int i = 0, xpos = 0; i < (*n); i++, xpos += (*xinc))  
  {  ans += (x[xpos] * x[xpos]);  }
  return(ans);
}


/********************************************************************
 * \sum wi (xi - x0)(yi - y0), 
 * where x is n-by-p matrix, y is n-by-1 vector
 * output is a p-by-1 vector
 * or simply t(x - x0) %*% diag(w) %*% (y - y0)
 ********************************************************************/

void sumwMatVecCtr(const int *n, const int *p,
     const double *x, const double *x0,
     const double *y, const double *y0,
     const double *w, double *wxy)
{
  for(int i = 0; i < (*p); i++) 
    wxy[i] = sum3Ctr(n, x + i * (*n), x0 + i, y, y0, w);
}


/********************************************************************
 * \sum wi xi xi^T, where x is n-by-p matrix 
 * output is a symmetric p-by-p matrix.  
 * or simply t(x) %*% diag(w) %*% x
 ********************************************************************/

void sumwMatMatCtr(const int *n, const int *p,
     const double *x, const double *x0,
     const double *w, double *wxx)
{
  int i, j;

  for(i = 0; i < (*p); i++)
  for(j = i; j < (*p); j++)
  {
    wxx[i + j * (*p)] = sum3Ctr(n, x + i * (*n), x0 + i, x + j * (*n), x0 + j, w);
    wxx[j + i * (*p)] = wxx[i + j * (*p)];
  }
}


/********************************************************************
 * scale the columns of x such that the column sum is colsum
 ********************************************************************/

void colScale(const int *n, const int *p, double *x, 
     const double *colsum)
{
  double sc, *xcol;
  int i, k;

  for(i = 0; i < (*p); i++)
  {
    xcol = x + i * (*n);
    sc = (*colsum) / sum(n, xcol);
    for(k = 0; k < (*n); k++) xcol[k] *= sc;
  }
}


/********************************************************************
 * find the value and position of max( abs(x) )
 ********************************************************************/

void maxabs(const int *p, const double *x, double *xmax, int *pos)
{
  double dtmp;

  (*xmax) = fabs(*x); (*pos) = 0;
  for(int i = 1; i < (*p); i++)
  {
    dtmp = fabs(x[i]);
    if(dtmp > (*xmax)) { (*xmax) = dtmp; (*pos) = i; }
  }  /*** return the value and position of max(abs(x)) ***/
}


/********************************************************************
 * find the value and position of min( x )
 ********************************************************************/

void findmin(const int *n, const double *x, int *pos, double *xmin)
{
  (*xmin) = (*x); (*pos) = 0;
  for(int i = 1; i < (*n); i++)
  {
    if(x[i] < (*xmin)) { (*xmin) = x[i]; (*pos) = i; }
  }  /*** return the value and position of min(x) ***/
}


/********************************************************************
 * find the value and position of max( x )
 ********************************************************************/

void findmax(const int *n, const double *x, int *pos, double *xmax)
{
  (*xmax) = (*x); (*pos) = 0;
  for(int i = 1; i < (*n); i++)
  {
    if(x[i] > (*xmax)) { (*xmax) = x[i]; (*pos) = i; }
  }  /*** return the value and position of max(x) ***/
}


/*****************************************************************
 * fill the upper-triangle elements of a symmetric matrix
 *****************************************************************/

void fill_sym_mat(const int *n, double *x)
{
  int i, j;
  for(i = 0; i < (*n); i++) 
  for(j = i + 1; j < (*n); j++)
    x[i + j * (*n)] = x[j + i * (*n)];
}


/*****************************************************************
 * linear interpolation
 * x1, x2, x0 are scalars, y1, y2, y0 are vectors
 *****************************************************************/

void interpolation(const int* p, 
     const double *x1, const double *y1, 
     const double *x2, const double *y2,
     const double *x0, double *y0)
{
  double ratio1 = ((*x2) - (*x0)) / ((*x2) - (*x1));
  double ratio2 = 1.0 - ratio1;
  for(int i = 0; i < (*p); i++) 
    y0[i] = ratio1 * y1[i] + ratio2 * y2[i];
}


/*****************************************************************
 * given a point one a line and assiciated slope
 * output the value of y at a new point
 *****************************************************************/

double linepred(const double *x0, const double *y0, 
     const double *slope, const double *x)
{
  return( (*y0) + (*slope) * ((*x) - (*x0)) );
}


/********************************************************************
 * lasso: the minimizer of
 *    x^T S x - 2 a^T x + 2 * lambda * sum(abs(x))
 ********************************************************************/

/********************************************************************
 * (varseq) is a permutation of {0, 1, ..., p-1} 
 *      which corresponds to p different variables
 * the first (nact) variables have nonzero coefficients
 *      which corresponds to those in the active set
 * the remaining variables have zero coefficients.
 *
 * (Scp) is a p-by-p matrix, and it contains two sub-matrices.
 * Scp[1:nact, 1:nact] is a lower-triangle matrix, which is the
 * Cholesky decomposition of Smat[varseq[1:nact], varseq[1:nact]]
 * Scp[(nact+1):p, 1:nact] = Smat[varseq[(nact+1):p], varseq[1:nact]]
 ********************************************************************/


/********************************************************************
 * the first step (find the minimum lambda that yields 0 solution)
 ********************************************************************/

void lasso_1st_step(
          const int* p, const double *Smat, const double* Avec,
          double* lambda, int* varseq, int* nact, int* action,
          double* Scp, double* beta)
{
  int i, pos, itmp;

  for(i = 0; i < (*p); i++) varseq[i] = i;
  for(i = 0; i < (*p); i++) beta[i] = 0.0;

  maxabs(p, Avec, lambda, &pos);
  (*nact) = 1; varseq[0] = pos; varseq[pos] = 0;
  (*action) = pos + 1;

  itmp = (*p) * pos;
  for(i = 1; i < (*p); i++)  Scp[i] = Smat[i + itmp];
  Scp[pos] = Smat[itmp];
  Scp[0] = sqrt(Smat[pos + itmp]);
}


/********************************************************************
 * add one variable varseq[pos] 
 * update the values of varseq, nact, Scp
 ********************************************************************/

void lasso_add(const int* p, const double* Smat, const int* pos,
               int* varseq, int* nact, double* Scp)
{
  int var, i, j, k;

  var = varseq[(*pos)];
  if((*pos) != (*nact))  /*** not 1st variable in inactive set ***/
  {
    varseq[(*pos)] = varseq[(*nact)];
    varseq[(*nact)] = var;
    F77_CALL(dcopy)(nact, Scp + (*nact), p, Scp + (*pos), p);
  }

  /*** add the new column to Scp[(nact+1):p, 1:nact]  ***/
  j = (*p) * (*nact); k = (*p) * var;
  for(i = (*nact) + 1; i < (*p); i++)
    Scp[i + j] = Smat[varseq[i] + k];

  /*** update Scp[1:nact, 1:nact], add variable "var" ***/
  for(i = 0; i < (*nact); i++)
    Scp[(*nact) + (*p) * i] = Smat[var + (*p) * varseq[i]];
  F77_CALL(dtrsm)("R", "L", "T", "N", &ONE, nact, &dONE, 
                  Scp, p, Scp + (*nact), p 
                  FCONE FCONE FCONE FCONE);

  Scp[(*nact) + (*nact) * (*p)] 
    = sqrt( Smat[var + var * (*p)] - sumsqx(nact, Scp + (*nact), p) );

  (*nact)++;
}


/*****************************************************************
 * remove one variable varseq[pos] 
 * update the values of varseq, nact, Scp
 *****************************************************************/

void lasso_remove(const int *p, const double *Smat, const int *pos,
                  int *varseq, int *nact, double *Scp)
{
  int var, i, j, k;

  (*nact)--;
  var = varseq[*pos];
  if((*pos) != (*nact)) /*** not last variable in inactive set ***/
  {
    varseq[(*pos)] = varseq[(*nact)];
    varseq[(*nact)] = var;
    i = (*nact) + 1; j = (*p) - (*nact);
    F77_CALL(dcopy)(&j, Scp + i + (*p) * (*nact), &ONE, 
                        Scp + i + (*p) * (*pos),  &ONE);

    /*** update Scp[1:nact, 1:nact], remove variable "var" ***/
    F77_CALL(dcopy)(pos, Scp + (*nact), p, Scp + (*pos), p);
    for(i = (*pos); i < (*nact); i++) for(j = (*pos); j <= i; j++)
      { Scp[i + j * (*p)] = Smat[varseq[i] + (*p) * varseq[j]]; } 

    i = (*pos) + (*pos) * (*p); k = (*nact) - (*pos);  
    F77_CALL(dsyrk)("L", "N", &k, pos, &dnONE, Scp + (*pos), p, 
                                       &dONE,  Scp + i,      p
                                       FCONE FCONE); 
    F77_CALL(dpotrf)("L", &k, Scp + i, p, &j FCONE);
  }

  /*** add the new row to Scp[(nact+1):p, 1:nact]  ***/
  for(i = 0; i < (*nact); i++)
    Scp[(*nact) + (*p) * i] = Smat[var + (*p) * varseq[i]];
}


/*****************************************************************
 * find smallest step length for removal
 * need to set an initial value of delta when calling it.
 * for example, set delta = lambda.
 *****************************************************************/

void lasso_steplength_remove(
          const int* nact, const int* varseq, const double* beta, 
          const double* Gvec, const double* eps,
          double* delta, int* pos)
{
  double dtmp;

  for(int i = 0; i < (*nact); i++)
  {
    dtmp = -beta[varseq[i]] / Gvec[i];
    if( (dtmp > (*eps)) && (dtmp < (*delta)) )
    { (*delta) = dtmp; (*pos) = i; }
  }
}


/*****************************************************************
 * find smallest step length for addition 
 * need to set an initial value of delta when calling it.
 * for example, set delta = lambda.
 *****************************************************************/

void lasso_steplength_add(
          const int* p, const int* nact, const int* varseq, 
          const double* Avec, const double* lambda,
          const double* SiG, const double* eps,
          double* delta, int* pos, int* add)
{
  const int ninact = (*p) - (*nact);
  double dtmp;

  for(int i = 0; i < ninact; i++)
  {
    dtmp = 1.0 + SiG[i];
    if( (dtmp > (*eps)) || (dtmp < -(*eps)) )
    {
      dtmp = ((*lambda) + Avec[varseq[(*nact) + i]]) / dtmp;
      if((dtmp > (*eps)) && (dtmp < (*delta))) 
      { (*delta) = dtmp; (*pos) = i + (*nact); (*add) = TRUE; }
    }

    dtmp = 1.0 - SiG[i];
    if( (dtmp > (*eps)) || (dtmp < -(*eps)) )
    {
      dtmp = ((*lambda) - Avec[varseq[(*nact) + i]]) / dtmp;
      if( (dtmp > (*eps)) && (dtmp < (*delta)) ) 
      { (*delta) = dtmp; (*pos) = i + (*nact); (*add) = TRUE; }
    }
  }
}



/*****************************************************************
 * update the value of beta, Avec, lambda
 *****************************************************************/

void lasso_update_beta(
          const int* p, const int* nact, const int* varseq,
          const double* delta, const double* Gvec, 
          const double* SiG, const double* Rsign,
          double* lambda, double* Avec, double* beta)
{
  int i;

  for(i = 0; i < (*nact); i++) 
    beta[varseq[i]] += ((*delta) * Gvec[i]);

  (*lambda) = fabs(Avec[*varseq]) - (*delta);

  for(i = 0; i < (*nact); i++) 
    Avec[varseq[i]] = (*lambda) * Rsign[i];
  for(i = (*nact); i < (*p); i++) 
    Avec[varseq[i]] -= ((*delta) * SiG[i - (*nact)]);
}


/*****************************************************************
 * one step in calculating lasso solution path
 * where xtx = t(x) %*% x is a symmetric matrix (p-by-p)
 *       xy  = t(x) %*% y is a vector (p-by-1)
 * there is no intercept term.
 *
 * lambda, old (input) and new (output) value of lambda. 
 * eps is an effective zero.
 * varseq is the sequence of integers {0, ..., p-1}, 
 * where the first nact elements indicate those in the active set.
 * action:(output) positive means addition, negative means removal.
 * beta: old (input) and new (output) value of beta.
 *****************************************************************/

void lasso_step(
          const int* p, const double* Smat, double* Avec,
          const double* eps, 
          double* lambda, int* varseq, int* nact, 
          int* action, double* Scp, double* beta,
          double* delta, double* Gvec, double* Rsign, double* SiG)
{ 
  int add, pos, info, ninact, i;

  /*** calculate direction Gvec ***/
  for(i = 0; i < (*nact); i++) 
  {
    Rsign[i] = sign(Avec[varseq[i]]);
    Gvec[i] = Rsign[i];
  }
  F77_CALL(dpotrs)("L", nact, &ONE, Scp, p, Gvec, p, &info FCONE);

  /*** possible removal ***/
  add = FALSE; (*delta) = (*lambda) + 1.0; pos = 0;
  lasso_steplength_remove(nact, varseq, beta, Gvec, eps, 
                          delta, &pos);

  /*** possible addition ***/
  if((*nact) < (*p))
  {
    ninact = (*p) - (*nact);
    F77_CALL(dgemv)("N", &ninact, nact, &dONE, Scp + (*nact), p, 
                         Gvec, &ONE, &dZERO, SiG, &ONE FCONE);
    lasso_steplength_add(p, nact, varseq, Avec, lambda, SiG, eps, 
                         delta, &pos, &add);
  }
 
  /*** move to lambda = 0, least squares estimate ***/
  if((*delta) >= (*lambda))
  {
    (*delta) = (*lambda); (*lambda) = 0.0; (*action) = 0;
    for(i = 0; i < (*nact); i++) 
      beta[varseq[i]] += ((*delta) * Gvec[i]);
  }
  else  /*** update beta, Avec, and varseq ***/
  {
    lasso_update_beta(p, nact, varseq, delta, Gvec, SiG, Rsign,
                      lambda, Avec, beta);

    if(add == TRUE)  
    {
      (*action) = varseq[pos] + 1;
      lasso_add(p, Smat, &pos, varseq, nact, Scp);
    }
    else /* remove variable */
    {
      (*action) = -(varseq[pos] + 1);
      beta[varseq[pos]] = 0.0; 
      lasso_remove(p, Smat, &pos, varseq, nact, Scp);
    }
  }
}
 

/*****************************************************************
 * calculate lasso solution for given value of lambda
 *****************************************************************/

void lasso_lambda(
          const int* p, const double* Smat, double* Avec, 
          const double* lda, const double *eps, double* beta)
{
  double *Gvec, *Rsign, *SiG, *beta0, *Scp;
  double lambda, lambda0, ratio, delta;
  int *varseq, nact, i, action;

  Gvec   = (double*)Calloc(*p, double);
  Rsign  = (double*)Calloc(*p, double);
  SiG    = (double*)Calloc(*p, double);
  beta0  = (double*)Calloc(*p, double);
  Scp    = (double*)Calloc((*p) * (*p), double);
  varseq = (int*)Calloc(*p, int);

  lasso_1st_step(p, Smat, Avec, &lambda, varseq, &nact, &action, Scp, beta);
  lambda0 = lambda;
  if( (*lda) >= lambda )
  { 
    Free(Gvec); Free(Rsign); Free(SiG); Free(beta0); Free(Scp); 
    Free(varseq);
    return;
  }

  while( lambda > (*lda) )
  {
    /*** save old values of beta and lambda ***/
    for(i = 0; i < (*p); i++) beta0[i] = beta[i]; 
    lambda0 = lambda; 

    lasso_step(p, Smat, Avec, eps, &lambda, varseq, &nact, &action, 
               Scp, beta, &delta, Gvec, Rsign, SiG);
  } /*** stop if lambda <= lda ***/

  /*** update value of beta ***/
  ratio = ((*lda) - lambda) / (lambda0 - lambda);
  for(i = 0; i < (*p); i++) 
    beta[i] = (1.0 - ratio) * beta[i] + ratio * beta0[i];

  Free(Gvec); Free(Rsign); Free(SiG); Free(beta0); Free(Scp); 
  Free(varseq);
}

 
/*****************************************************************
 * weighted least squares with lasso penalty 
 *****************************************************************/

void wlasso0(const int *n, const int *p, 
          const double *y, const double *x, const double *w,
          const double *lambda, 
          double *ymean, double *xmean,
          double *xwx, double *xwy, double *beta)
{
  const double eps = DOUBLE_EPS;
  double wsum;

  wsum = sum(n, w);
  (*ymean) = sum2(n, y, w) / wsum;
  for(int k = 0; k < (*p); k++) 
    xmean[k] = sum2(n, x + k * (*n), w) / wsum;

  sumwMatVecCtr(n, p, x, xmean, y, ymean, w, xwy);
  sumwMatMatCtr(n, p, x, xmean, w, xwx);
  lasso_lambda(p, xwx, xwy, lambda, &eps, beta);
}


/*****************************************************************
 * weighted least squares with lasso penalty, univariate x
 *****************************************************************/

void unilasso(const int *n, 
          const double *y, const double *x, const double *w, 
          const double *lambda, double *a, double *b)
{
  if((*lambda) < 0.0) { (*a) = 0.0; (*b) = 0.0; return; }

  double xwx, xwy, wsum, xmu, ymu, Rabs;

  wsum = sum(n, w);
  xmu  = sum2(n, x, w) / wsum;
  ymu  = sum2(n, y, w) / wsum;

  xwx  = sum3Ctr(n, x, &xmu, x, &xmu, w);
  xwy  = sum3Ctr(n, x, &xmu, y, &ymu, w);

  Rabs = fabs(xwy);

  if((*lambda) >= Rabs) (*b) = 0.0; 
  else (*b) = (1.0 - (*lambda) / Rabs) * xwy / xwx;

  (*a) = R_FINITE(*b) ? (ymu - (*b) * xmu) : (ymu); 
}


/********************************************************************
 * OPGlasso
 ********************************************************************/

void calwvec(const int *n, const int *p,
     const double *x, const double *h,
     const double *x0, const int *xinc, double *w)
{
  const double sc = -0.5 / (*h) / (*h);
  double temp, len2;
  int i, k;

  for(i = 0; i < (*n); i++)
  {
    len2 = 0.0;
    for(k = 0; k < (*p); k++) 
    { temp = x[i + k * (*n)] - x0[k * (*xinc)]; len2 += temp * temp; }

    w[i] = exp(len2 * sc);
  }
}


void OPGlasso(const int *n, const int *p, 
     const double *y, const double *x, 
     const double *h, const double *lambda,
     double *betamat)
{
  double *w, *xwx, *xwy;
  double y0, *x0;
  int i, k;

  x0  = (double*)Calloc(*p, double);
  w   = (double*)Calloc(*n, double);
  xwx = (double*)Calloc((*p) * (*p), double);
  xwy = (double*)Calloc(*p, double);

  for(i = 0; i < (*n); i++)
  {
    calwvec(n, p, x, h, x + i, n, w);
    w[i] = 0.0;  /* exclude ith point itself */
    /* scale w such that sum(w) = 1.0 */
    y0 = 1.0 / sum(n, w);
    for(k = 0; k < (*n); k++) w[k] *= y0;
    wlasso0(n, p, y, x, w, lambda, &y0, x0, xwx, xwy, betamat + i * (*p));
  }

  Free(x0); Free(w); Free(xwx); Free(xwy);
}


/********************************************************************
 * (univariate) weighted least squares estimate
 * lm(y ~ x, weights = w)
 ********************************************************************/

void uniwlse(const int *n, const double *y, const double *x, 
             const double *w, double *a, double *b)
{
  double xwx, xwy, wsum, xmu, ymu;

  wsum = sum(n, w);
  xmu  = sum2(n, x, w) / wsum;
  ymu  = sum2(n, y, w) / wsum;

  xwx  = sum3Ctr(n, x, &xmu, x, &xmu, w);
  xwy  = sum3Ctr(n, x, &xmu, y, &ymu, w);

  (*b) = xwy / xwx;
  (*a) = R_FINITE(*b) ? (ymu - (*b) * xmu) : (ymu);  
}


/*****************************************************************
 * calculate solve(xwx, xwy)
 * xwx is a p-by-p matrix definite positive matrix
 * xwy is a p-by-q matrix
 *
 * least squares estimate using Cholesky decomposition
 * xwx = t(x) %*% w %*% x;  xwy = t(x) %*% w %*% y
 *****************************************************************/

void solve(const int *p, const int *q, double *xwx, double *xwy)
{
  int info;

  F77_CALL(dpotrf)("L", p, xwx, p, &info FCONE);
  F77_CALL(dpotrs)("L", p, q, xwx, p, xwy, p, &info FCONE);
}


/*****************************************************************
 * calculate a symmetric n-by-n weight matrix (wmat)
 * Gaussian kernel function
 * x is an n-by-1 vector and h is the bandwidth.
 * diag == TRUE   diagonal elements are ones.
 * diag == FALSE  diagonal elements are zeros.
 *****************************************************************/

void uniweight(const int *n, const double *x, const double *h, 
               const int *diag, double *wmat)
{
  const int nn = (*n) * (*n), n1 = (*n) + 1;
  const double sc = (-0.5) / ((*h) * (*h));
  double d0, d1;
  int i, j;

  /*** off-diagonal elements ***/
  for(i = 0; i < (*n); i++)
    for(j = 0; j < i; j++)
    {
      d0 = x[i] - x[j];
      d1 = exp(d0 * d0 * sc);
      wmat[i + j * (*n)] = d1;
      wmat[j + i * (*n)] = d1;
    }

  /*** diagonal elements ***/
  d0 = (*diag) ? (1.0) : (0.0);
  for(i = 0; i < nn; i += n1) wmat[i] = d0;
}
  

/**********************************************************************
 * kernel regression. 
 * evaluate m(x) at newx[k]
 **********************************************************************/

void kernelreg(const int *n, const int *p,
               const double *y, const double *x, const double *h,
               const double *wt, const int *m, const double *newx,
               double *yhat)
{
  const double sc = -0.5 / (*h) / (*h);
  double yw, wsum, len2, tmp;
  int i, j, k;

  for(k = 0; k < (*m); k++)
  {
    yw = 0.0; wsum = 0.0;
    for(i = 0; i < (*n); i++)
    {
      len2 = 0.0;
      for(j = 0; j < (*p); j++)
      {  
        tmp = x[i + j * (*n)] - newx[k + j * (*m)];  
        len2 += (tmp * tmp);
      }
      tmp = exp(len2 * sc) * wt[i];
      yw += (y[i] * tmp); wsum += tmp;
    }

    yhat[k] = yw / wsum;
  } /* evaluate m(x) at each component of newx */
}


/**********************************************************************
 * univariate local linear smoothing. 
 * evaluate m(x) at newx[k]
 * a and b are m-by-1 vectors
 **********************************************************************/

void uniloclin(const int *n,  
               const double *y, const double *x, const double *h, 
               const double *wt, const int *m, const double *newx,
               double *a, double *b)
{
  const double sc = -0.5 / (*h) / (*h);
  double *xij, *w;
  double tmp;
  int i, k;

  xij = (double*)Calloc((*n), double);
  w   = (double*)Calloc((*n), double);

  for(k = 0; k < (*m); k++)
  {
    for(i = 0; i < (*n); i++)
    {
      tmp = x[i] - newx[k];
      xij[i] = tmp;
      w[i] = exp(tmp * tmp * sc) * wt[i];
    }

    uniwlse(n, y, xij, w, a + k, b + k);
  } /* evaluate m(x) at each component of newx */

  Free(xij); Free(w);
}


/**********************************************************************
 * if newx is beyond the range of x, do linear extraploation
 * using the slope at min(x) or max(x)
 **********************************************************************/

void uniloclin_exp(const int *n,  
               const double *y, const double *x, const double *h, 
               const double *wt, const int *m, const double *newx,
               double *a, double *b)
{
  double xr[2], ar[2], br[2], xcenter;
  int pos;

  findmin(n, x, &pos, xr);
  findmax(n, x, &pos, xr + 1);
  pos = 2;
  uniloclin(n, y, x, h, wt, &pos, xr, ar, br);

  xcenter = (xr[0] + xr[1]) / 2.0;
  for(int k = 0; k < (*m); k++)
  {
    if(newx[k] <= xcenter) pos = 0;  /** left **/
    else                   pos = 1;  /** right **/

    a[k] = linepred(xr + pos, ar + pos, br + pos, newx + k);
    b[k] = br[pos];
  } /* evaluate m(x) at each component of newx */
}


/********************************************************************
 * xwx = \sum w[i,j] * b[j]^2 (x[i, ] - x[j, ]) * (x[i, ] - x[j, ])^T
 * xwy = \sum w[i,j] * b[j] * (y[i] - a[j]) * (x[i, ] - x[j, ])
 ********************************************************************/

void cal_xwx_xwy(const int *n, const int *p,
     const double *y, const double *x, const double *wmat,
     const double *a, const double *b,  
     double *xwx, double *xwy)
{
  const int ONE = 1;
  const int pp = (*p) * (*p);
  double bi, bibi, bj, w0, *xij;
  int i, j, k;

  xij = (double*)Calloc(*p, double);
 
  for(i = 0; i <  pp;  i++)  xwx[i] = 0.0;
  for(i = 0; i < (*p); i++)  xwy[i] = 0.0;

  for(i = 0; i < (*n); i++)
  {
    if( R_FINITE(b[i]) ) { bi = b[i]; bibi = bi * bi; }
    else { bi = 0.0; bibi = 0.0; }
    /* skip if b[i] = Inf, -Inf, NA, NaN */

    for(j = 0; j < i; j++)
    {
      if( R_FINITE(b[j]) ) bj = b[j];  
      else bj = 0.0;
      /* skip if b[j] = Inf, -Inf, NA, NaN */

      for(k = 0; k < (*p); k++) 
        xij[k] = x[i + k * (*n)] - x[j + k * (*n)];

      w0 = wmat[j + i * (*n)] * bibi + wmat[i + j * (*n)] * bj * bj;
      F77_CALL(dsyr)("L", p, &w0, xij, &ONE, xwx, p FCONE);

      w0 = wmat[i + j * (*n)] * (y[i] - a[j]) * bj - wmat[j + i * (*n)] * (y[j] - a[i]) * bi;
      F77_CALL(daxpy)(p, &w0, xij, &ONE, xwy, &ONE);
    }
  }

  fill_sym_mat(p, xwx);
  Free(xij);
}


/********************************************************************
 * one iteration step for MAVE in single-index models
 ********************************************************************/

void sim_beta2ab(const int *n, const int *p,
     const double *y, const double *xb0, const double *wmat,
     double *a, double *b)
{
  double *xij;
  int j, k;

  xij  = (double*)Calloc(*n, double);

  for(j = 0; j < (*n); j++)
  {
    for(k = 0; k < (*n); k++)  xij[k] = xb0[k] - xb0[j];  
    uniwlse(n, y, xij, wmat + j * (*n), a + j, b + j);
  }

  Free(xij);
}


void sim_ab2beta(const int *n, const int *p,
     const double *y, const double *x, const double *w,
     const double *a, const double *b, double *beta)
{
  double *xwx;

  xwx  = (double*)Calloc((*p) * (*p), double);

  cal_xwx_xwy(n, p, y, x, w, a, b, xwx, beta);
  solve(p, &ONE, xwx, beta);

  Free(xwx);  
}


void sim_update(const int *n, const int *p,
     const double *y, const double *x, const double *h,
     const double *beta0, const int *diagin,
     double *a, double *b, double *beta)
{
  double *xb0, *wmat;
  xb0  = (double*)Calloc(*n, double);
  wmat = (double*)Calloc((*n)*(*n), double);

  F77_CALL(dgemv)("N", n, p, &dONE, x, n, beta0, &ONE, &dZERO, xb0, &ONE FCONE);
  uniweight(n, xb0, h, diagin, wmat);
  colScale(n, n, wmat, &dONE);

  sim_beta2ab(n, p, y, xb0, wmat, a, b);
  sim_ab2beta(n, p, y, x, wmat, a, b, beta);

  Free(xb0); Free(wmat);
}


/********************************************************************
 * one iteration step for sim-lasso
 ********************************************************************/
 
void simlasso_beta2ab(const int *n, const int *p,
     const double *y, const double *xb0, const double *wmat,
     const double *lambda, double *a, double *b)
{
  double *xij;
  int j, k;

  xij  = (double*)Calloc(*n, double);

  for(j = 0; j < (*n); j++)
  {
    for(k = 0; k < (*n); k++)  xij[k] = xb0[k] - xb0[j];  
    unilasso(n, y, xij, wmat + j * (*n), lambda, a + j, b + j);
  }

  Free(xij);
}


void simlasso_ab2beta(const int *n, const int *p,
     const double *y, const double *x, const double *w,
     const double *lambda, const double *a, const double *b,
     double *beta)
{
  const double eps = DOUBLE_EPS;
  double *xwx, *xwy, adjlda;

  xwx  = (double*)Calloc((*p) * (*p), double);
  xwy  = (double*)Calloc(*p, double);

  cal_xwx_xwy(n, p, y, x, w, a, b, xwx, xwy);

  adjlda = sumabs(n, b);
  if(adjlda > eps)  /*** continue only if adjlda > 0 ***/
  {
    adjlda *= (*lambda);  /*** adjlda = lambda * sum(abs(b)) ***/
    lasso_lambda(p, xwx, xwy, &adjlda, &eps, beta);
  }
  else for(int k = 0; k < (*p); k++) beta[k] = 0.0;
  /*** adjlda == 0 means all b[j] == 0 ***/

  Free(xwx); Free(xwy);
}


void simlasso_update(const int *n, const int *p,
     const double *y, const double *x, const double *h,
     const double *lambda, const double *beta0, const int *diagin,
     double *a, double *b, double *beta)
{
  double *xb0, *wmat, adjlda;

  xb0  = (double*)Calloc(*n, double);
  wmat = (double*)Calloc((*n) * (*n), double);

  F77_CALL(dgemv)("N", n, p, &dONE, x, n, beta0, &ONE, &dZERO, xb0, &ONE FCONE);
  uniweight(n, xb0, h, diagin, wmat);
  colScale(n, n, wmat, &dONE);

  /*** adjlda = lambda * sum(abs(beta0)) ***/
  adjlda = sumabs(p, beta0) * (*lambda);   

  simlasso_beta2ab(n, p, y, xb0, wmat, &adjlda, a, b);
  simlasso_ab2beta(n, p, y, x, wmat, lambda, a, b, beta);

  Free(xb0); Free(wmat);
}


/********************************************************************
 * one iteration step for sMAVE (algorithm in Wang and Yin (2008))
 ********************************************************************/

void sMAVE_ab2beta(const int *n, const int *p,
     const double *y, const double *x, const double *w,
     const double *lambda, const double *a, const double *b,
     double *beta)
{
  const double eps = DOUBLE_EPS;
  double *xwx, *xwy;

  xwx  = (double*)Calloc((*p) * (*p), double);
  xwy  = (double*)Calloc(*p, double);

  cal_xwx_xwy(n, p, y, x, w, a, b, xwx, xwy);
  lasso_lambda(p, xwx, xwy, lambda, &eps, beta);

  Free(xwx); Free(xwy);
}


void sMAVE_update(const int *n, const int *p,
     const double *y, const double *x, const double *h,
     const double *lambda, const double *beta0, const int *diagin,
     double *a, double *b, double *beta)
{
  double *xb0, *wmat;

  xb0  = (double*)Calloc(*n, double);
  wmat = (double*)Calloc((*n) * (*n), double);

  F77_CALL(dgemv)("N", n, p, &dONE, x, n, beta0, &ONE, &dZERO, xb0, &ONE FCONE);
  uniweight(n, xb0, h, diagin, wmat);
  colScale(n, n, wmat, &dONE);

  sim_beta2ab(n, p, y, xb0, wmat, a, b);
  sMAVE_ab2beta(n, p, y, x, wmat, lambda, a, b, beta);

  Free(xb0); Free(wmat);
}


/**********************************************************************/
/*****                      END OF THE FILE                       *****/
/**********************************************************************/
