######################################################################
## sim-lasso
######################################################################

# dyn.load("sim-lasso-clib.dll")

######################################################################

OPG.lasso = function(y, x, h, lambda)
{
  check.argument(y, x, h, lambda)
  n = NROW(x); p = NCOL(x);

  bmat = .C("OPGlasso", as.integer(n), as.integer(p), 
     as.double(y), as.double(x), as.double(h), as.double(lambda),
     bmat = double(p * n))$bmat
  bmat = matrix(bmat, nrow = n, ncol = p, byrow = T)
  beigen = eigen(t(bmat) %*% bmat, symmetric = T) 

  list(method = "OPG-lasso", h = h, lambda = lambda,
       bmat = bmat, beigen = beigen, 
       beta = beigen$vectors[, 1])
}


######################################################################

sim = function(y, x, h, b0 = NULL, 
               niter = 50, tol = 0.001, diag.in = FALSE)
{
  check.argument(y, x, h, 0)
  n = NROW(x); p = NCOL(x)

  b0fit = initial.simlasso(y, x, h, 0, b0) 
  newbeta = b0fit$beta

  conv = tol + 1; iter = 0;
  while((iter < niter) & (conv > tol) & (!is.na(conv)) )
  {
    oldbeta = newbeta
    fit = C.sim.update(y, x, h, oldbeta, diag.in)

    if(fit$betalen != 0) newbeta = fit$beta / fit$betalen
    else newbeta = fit$beta

    iter = iter + 1
    conv = dist.vec(oldbeta, newbeta)
  }

  list(method = "sim", h = h, b0fit = b0fit,
       beta = newbeta, betalen = fit$betalen,
       a = fit$a, b = fit$b, iter = iter, conv = conv)
}


C.sim.update = function(y, x, h, b0, diag.in)
{
  n = NROW(x); p = NCOL(x)
  ans = .C("sim_update", as.integer(n), as.integer(p),
           as.double(y), as.double(x), as.double(h),
           as.double(b0), as.integer(diag.in),
           a = double(n), b = double(n), beta = double(p))
  betalen = sqrt(sum(ans$beta * ans$beta)) 
  list(beta = ans$beta, a = ans$a, b = ans$b, betalen = betalen)
}


######################################################################

sim.lasso = function(y, x, h, lambda, b0 = NULL, 
                     niter = 50, tol = 0.001, diag.in = FALSE)
{
  check.argument(y, x, h, lambda)
  n = NROW(x); p = NCOL(x)

  b0fit = initial.simlasso(y, x, h, lambda, b0) 
  newbeta = b0fit$beta

  conv = tol + 1; iter = 0;
  while((iter < niter) & (conv > tol) & (!is.na(conv)) )
  {
    oldbeta = newbeta
    fit = C.sim.lasso.update(y, x, h, lambda, oldbeta, diag.in)

    if(fit$betalen != 0) newbeta = fit$beta / fit$betalen
    else newbeta = fit$beta

    iter = iter + 1
    conv = dist.vec(oldbeta, newbeta)
  }

  list(method = "sim-lasso", h = h, lambda = lambda, b0fit = b0fit,
       beta = newbeta, betalen = fit$betalen,
       a = fit$a, b = fit$b, iter = iter, conv = conv)
}


C.sim.lasso.update = function(y, x, h, lambda, b0, diag.in)
{
  n = NROW(x); p = NCOL(x)
  ans = .C("simlasso_update", 
        as.integer(n), as.integer(p), as.double(y),
        as.double(x), as.double(h), as.double(lambda),
        as.double(b0), as.integer(diag.in),
        a = double(n), b = double(n), beta = double(p))
  betalen = sqrt(sum(ans$beta * ans$beta)) 
  list(beta = ans$beta, a = ans$a, b = ans$b, betalen = betalen)
}


sim.lasso.path = function(y, x, h, lambda, b0 = NULL, 
                     niter = 50, tol = 0.001, diag.in = FALSE)
{
  n = NROW(x); p = NCOL(x); LL = length(lambda)
  betamat = matrix(0, nrow = LL, ncol = p)
  rownames(betamat) = paste(lambda)
  bjinfo  = matrix(0, nrow = LL, ncol = 2)
  colnames(bjinfo) = c("norm1", "zeros")
  rownames(bjinfo) = paste(lambda)

  ans1 = sim.lasso(y, x, h, lambda[1], b0, niter, tol, diag.in)
  betamat[1, ] = ans1$beta
  bjinfo[1, 1] = sum(abs(ans1$b)); bjinfo[1, 2] = sum(ans1$b == 0)
  if(LL >= 2)
  {
    b0x = betamat[1, ]
    ans2 = sim.lasso(y, x, h, lambda[2], b0x, niter, tol, diag.in)
    betamat[2, ] = ans2$beta
    bjinfo[2, 1] = sum(abs(ans2$b)); bjinfo[2, 2] = sum(ans2$b == 0)
  }

  if(LL >= 3) for(i in 3:LL)
  {
    b0x = 2 * betamat[i - 1, ] - betamat[i - 2, ]
    ans = sim.lasso(y, x, h, lambda[i], b0x, niter, tol, diag.in)
    betamat[i, ] = ans$beta
    bjinfo[i, 1] = sum(abs(ans$b)); bjinfo[i, 2] = sum(ans$b == 0)
  }

  list(h = h, lambda = lambda, beta = betamat, bj = bjinfo)
}


######################################################################

sim.sMAVE = function(y, x, h, lambda, b0 = NULL, 
                     niter = 50, tol = 0.001, diag.in = FALSE)
{
  check.argument(y, x, h, lambda)
  n = NROW(x); p = NCOL(x)

  b0fit = initial.sMAVE(y, x, h, lambda, b0) 
  newbeta = b0fit$beta

  conv = tol + 1; iter = 0;
  while((iter < niter) & (conv > tol) & (!is.na(conv)) )
  {
    oldbeta = newbeta
    fit = C.sim.sMAVE.update(y, x, h, lambda, oldbeta, diag.in)    

    if(fit$betalen != 0) newbeta = fit$beta / fit$betalen
    else newbeta = fit$beta

    iter = iter + 1
    conv = dist.vec(oldbeta, newbeta)
  }

  list(method = "sim-sMAVE", h = h, lambda = lambda, b0fit = b0fit,
       beta = newbeta, betalen = fit$betalen,
       a = fit$a, b = fit$b, iter = iter, conv = conv)
}
 

C.sim.sMAVE.update = function(y, x, h, lambda, b0, diag.in)
{
  n = NROW(x); p = NCOL(x)
  ans = .C("sMAVE_update", 
        as.integer(n), as.integer(p), as.double(y),
        as.double(x), as.double(h), as.double(lambda),
        as.double(b0), as.integer(diag.in),
        a = double(n), b = double(n), beta = double(p))
  betalen = sqrt(sum(ans$beta * ans$beta)) 
  list(beta = ans$beta, a = ans$a, b = ans$b, betalen = betalen)
}


######################################################################

initial.simlasso = function(y, x, h, lambda, b0) 
{
  n = NROW(x); p = NCOL(x);

  b0len = sqrt( sum(b0 * b0) )
  if(is.null(b0) | (b0len == 0))
  {
    h0 = choose.h(x, 2)
    fit = OPG.lasso(y, x, h0, lambda)
    b0std = fit$beta
    inimethod = "OPG.lasso"
  }
  else if(length(b0) != p)
  {
    b0std = coef(lm(y ~ x))[-1]
    inimethod = "LSE"
  }
  else 
  { 
    b0std = b0 
    inimethod = "given"
  }

  b0std = b0std / sqrt( sum(b0std * b0std) )
  list(beta = b0std, method = inimethod)
}
 

initial.sMAVE = function(y, x, h, lambda, b0) 
{
  n = NROW(x); p = NCOL(x);

  b0len = sqrt( sum(b0 * b0) )
  if((length(b0) != p) | (b0len == 0))
  {
    b0std = coef(lm(y ~ x))[-1]
    inimethod = "LSE"
  }
  else 
  { 
    b0std = b0 
    inimethod = "given"
  }

  b0std = b0std / sqrt( sum(b0std * b0std) )
  list(beta = b0std, method = inimethod)
}


check.argument = function(y, x, h, lambda)
{
  n = NROW(x); p = NCOL(x);
  if(h < 0.0) stop("h < 0.0.\n")
  if(lambda < 0.0) stop("lambda < 0.0.\n")
  if(length(y) != n) stop("length(y) != NROW(x).\n")
}


######################################################################
## angle (in degree) between two vectors
## return NA if one vector is zero
######################################################################

dist.vec = function(a, b)
{
  ablen = sqrt(sum(a * a) * sum(b * b))
  if(is.na(ablen) || (ablen == 0)) ans = NA
  else ans = acos( abs(sum(a * b)) / ablen ) * 180 / pi
  ans
}

######################################################################
## misc
######################################################################

commondir = function(bmat)
{
  rowlength = sqrt( rowSums(bmat * bmat) )
  bmat0 = bmat / rowlength
  beigen = eigen(t(bmat0) %*% bmat0, symmetric = T) 
  list(bmat0 = bmat0, beigen = beigen, beta = beigen$vectors[, 1])
}


######################################################################
## rule of thumb
######################################################################

choose.h = function(x, sc = 2)
{
  n = NROW(x); p = NCOL(x)
  (4 / (2 * p + 1) / n)^(1/(p + 4)) * median(sd(x)) * sc
}


######################################################################
## more functions
######################################################################

prediction = function(y, x, h, theta, xnew = x)
{
  n = NROW(x); p = NCOL(x); m = NROW(xnew)
  if(length(y) != n)  stop("length(y) != NROW(x).\n")
  if(NCOL(xnew) != p) stop("NCOL(xnew) != NCOL(x).\n")

  if( any(is.na(theta)) || all(theta == 0) ) ### kernel estimate
  {
    yhat = .C("kernelreg", as.integer(n), as.integer(p),
              as.double(y), as.double(x), as.double(h),
              as.double(rep(1, n)), as.integer(m), as.double(xnew),
              yhat = double(m))$yhat
    ans = list(yhat = yhat, b = NULL)
  }
  else  ### local linear estimate
  {
    xb = x %*% theta; xbnew = xnew %*% theta 
    ans = loclin.pred(y, xb, h, rep(1, length(x)), xbnew)
  }

  ans   ## list(yhat = , b = )
}


loclin.pred = function(y, x, h, w = rep(1, length(x)), newx = x)
{
  n = length(y); m = length(newx)
  if(length(x) != n) stop("length(x) != length(y).\n")

  index = (newx <= max(x)) & (newx >= min(x))
  yhat = numeric(m); b = numeric(m);

  m1 = sum(index)
  if(m1 >= 1)  ## within the range
  {
    ans = .C("uniloclin", as.integer(n), as.double(y), as.double(x),
            as.double(h), as.double(w), as.integer(m1), 
            as.double(newx[index]), a = double(m1), b = double(m1))
    yhat[index] = ans$a; b[index] = ans$b
  }

  m2 = sum(!index)
  if(m2 >= 1)  ## beyond the range
  {
    ans = .C("uniloclin_exp", as.integer(n), as.double(y), as.double(x), 
            as.double(h), as.double(w), as.integer(m2), 
            as.double(newx[!index]), a = double(m2), b = double(m2))
   yhat[!index] = ans$a; b[!index] = ans$b
  }

  list(yhat = yhat, b = b)
}


CVm = function(y, x, h, lambda, m = 10, ...)
{
  n = NROW(x); p = NCOL(x);  
  if(length(y) != n) stop("length(y) != NROW(x).\n")

  index = sample( ( (1:n) %% m ) + 1 )
  cvmat = numeric(m) 
  for(i in 1:m)
  {
    ytrain = y[index != i]; xtrain = x[index != i, ]
    fit = sim.lasso(ytrain, xtrain, h, lambda, ...)
    yvalid = y[index == i]; xvalid = x[index == i, ]

    ans = prediction(ytrain, xtrain, h, fit$beta, xvalid)
    cvmat[i] = mean( (yvalid - ans$yhat)^2 )
  }

  list(h = h, lambda = lambda, cvmat = cvmat, cv = mean(cvmat))
}


choose.lambda = function(y, x, h, lda.seq, m = 10, ...)
{
  n = NROW(x); p = NCOL(x);  
  if(length(y) != n) stop("length(y) != NROW(x).\n")

  LL = length(lda.seq)
  cvmat = matrix(0, nrow = m, ncol = LL)
  for(i in 1:LL)
  {
    ans = CVm(y, x, h, lda.seq[i], m = m, ...)
    cvmat[, i] = ans$cvmat
  }
  cvmean = colMeans(cvmat) 
  cvsd   = sd(cvmat)

  pos1 = which.min(cvmean)
  pos2 = max(which(cvmean < cvmean[pos1] + cvsd[pos1] / sqrt(m)))

  list(cvmat = cvmat, cvmean = cvmean, cvsd = cvsd,
       opt = lda.seq[pos1], upper = lda.seq[pos2])
}


######################################################################
## END OF THE FILE
######################################################################
